@@ -1,13 +1,14 @@
 from __future__ import absolute_import
 
-import warnings
 from importlib import import_module
 
 from django import forms
+from django.contrib.auth import password_validation
 from django.contrib.auth.tokens import PasswordResetTokenGenerator
 from django.contrib.sites.shortcuts import get_current_site
 from django.core import exceptions, validators
 from django.urls import reverse
+from django.utils.safestring import mark_safe
 from django.utils.translation import gettext, gettext_lazy as _, pgettext
 
 from ..utils import (
@@ -20,6 +21,7 @@ from .adapter import get_adapter
 from .app_settings import AuthenticationMethod
 from .models import EmailAddress
 from .utils import (
+    assess_unique_email,
     filter_users_by_email,
     get_user_model,
     perform_login,
@@ -47,7 +49,7 @@ class EmailAwarePasswordResetTokenGenerator(PasswordResetTokenGenerator):
         return ret
 
 
-default_token_generator = EmailAwarePasswordResetTokenGenerator()
+default_token_generator = app_settings.PASSWORD_RESET_TOKEN_GENERATOR()
 
 
 class PasswordVerificationMixin(object):
@@ -88,7 +90,6 @@ class SetPasswordField(PasswordField):
 
 
 class LoginForm(forms.Form):
-
     password = PasswordField(label=_("Password"), autocomplete="current-password")
     remember = forms.BooleanField(label=_("Remember Me"), required=False)
 
@@ -96,7 +97,7 @@ class LoginForm(forms.Form):
     error_messages = {
         "account_inactive": _("This account is currently inactive."),
         "email_password_mismatch": _(
-            "The e-mail address and/or password you specified are not correct."
+            "The email address and/or password you specified are not correct."
         ),
         "username_password_mismatch": _(
             "The username and/or password you specified are not correct."
@@ -110,11 +111,11 @@ class LoginForm(forms.Form):
             login_widget = forms.TextInput(
                 attrs={
                     "type": "email",
-                    "placeholder": _("E-mail address"),
+                    "placeholder": _("Email address"),
                     "autocomplete": "email",
                 }
             )
-            login_field = forms.EmailField(label=_("E-mail"), widget=login_widget)
+            login_field = forms.EmailField(label=_("Email"), widget=login_widget)
         elif app_settings.AUTHENTICATION_METHOD == AuthenticationMethod.USERNAME:
             login_widget = forms.TextInput(
                 attrs={"placeholder": _("Username"), "autocomplete": "username"}
@@ -130,7 +131,7 @@ class LoginForm(forms.Form):
                 == AuthenticationMethod.USERNAME_EMAIL
             )
             login_widget = forms.TextInput(
-                attrs={"placeholder": _("Username or e-mail"), "autocomplete": "email"}
+                attrs={"placeholder": _("Username or email"), "autocomplete": "email"}
             )
             login_field = forms.CharField(
                 label=pgettext("field label", "Login"), widget=login_widget
@@ -139,6 +140,11 @@ class LoginForm(forms.Form):
         set_form_field_order(self, ["login", "password", "remember"])
         if app_settings.SESSION_REMEMBER is not None:
             del self.fields["remember"]
+        self.fields["password"].help_text = mark_safe(
+            _('<a href="{}">Forgot your password?</a>').format(
+                reverse("account_reset_password")
+            )
+        )
 
     def user_credentials(self):
         """
@@ -251,16 +257,10 @@ def _base_signup_form_class():
             'Module "%s" does not define a' ' "%s" class' % (fc_module, fc_classname)
         )
     if not hasattr(fc_class, "signup"):
-        if hasattr(fc_class, "save"):
-            warnings.warn(
-                "The custom signup form must offer"
-                " a `def signup(self, request, user)` method",
-                DeprecationWarning,
-            )
-        else:
-            raise exceptions.ImproperlyConfigured(
-                'The custom signup form must implement a "signup" method'
-            )
+        raise exceptions.ImproperlyConfigured(
+            "The custom signup form must offer"
+            " a `def signup(self, request, user)` method",
+        )
     return fc_class
 
 
@@ -276,7 +276,7 @@ class BaseSignupForm(_base_signup_form_class()):
         widget=forms.TextInput(
             attrs={
                 "type": "email",
-                "placeholder": _("E-mail address"),
+                "placeholder": _("Email address"),
                 "autocomplete": "email",
             }
         )
@@ -287,13 +287,6 @@ class BaseSignupForm(_base_signup_form_class()):
         self.username_required = kwargs.pop(
             "username_required", app_settings.USERNAME_REQUIRED
         )
-        # We can only truly prevent enumeration if email verification is
-        # mandatory, because in that case, regardless of whether or not the
-        # email is in use, the user will always receive an email.
-        self.prevent_enumeration = app_settings.PREVENT_ENUMERATION and (
-            app_settings.EMAIL_VERIFICATION
-            == app_settings.EmailVerificationMethod.MANDATORY
-        )
         self.account_already_exists = False
         super(BaseSignupForm, self).__init__(*args, **kwargs)
         username_field = self.fields["username"]
@@ -312,19 +305,19 @@ class BaseSignupForm(_base_signup_form_class()):
         ]
         if app_settings.SIGNUP_EMAIL_ENTER_TWICE:
             self.fields["email2"] = forms.EmailField(
-                label=_("E-mail (again)"),
+                label=_("Email (again)"),
                 widget=forms.TextInput(
                     attrs={
                         "type": "email",
-                        "placeholder": _("E-mail address confirmation"),
+                        "placeholder": _("Email address confirmation"),
                     }
                 ),
             )
         if email_required:
-            self.fields["email"].label = gettext("E-mail")
+            self.fields["email"].label = gettext("Email")
             self.fields["email"].required = True
         else:
-            self.fields["email"].label = gettext("E-mail (optional)")
+            self.fields["email"].label = gettext("Email (optional)")
             self.fields["email"].required = False
             self.fields["email"].widget.is_required = False
             if self.username_required:
@@ -356,16 +349,22 @@ class BaseSignupForm(_base_signup_form_class()):
         value = self.cleaned_data["email"]
         value = get_adapter().clean_email(value)
         if value and app_settings.UNIQUE_EMAIL:
-            try:
-                value = self.validate_unique_email(value)
-            except forms.ValidationError:
-                if not self.prevent_enumeration:
-                    raise
-                self.account_already_exists = True
+            value = self.validate_unique_email(value)
         return value
 
     def validate_unique_email(self, value):
-        return get_adapter().validate_unique_email(value)
+        adapter = get_adapter()
+        assessment = assess_unique_email(value)
+        if assessment is True:
+            # All good.
+            pass
+        elif assessment is False:
+            # Fail right away.
+            raise forms.ValidationError(adapter.error_messages["email_taken"])
+        else:
+            assert assessment is None
+            self.account_already_exists = True
+        return adapter.validate_unique_email(value)
 
     def clean(self):
         cleaned_data = super(BaseSignupForm, self).clean()
@@ -377,25 +376,34 @@ class BaseSignupForm(_base_signup_form_class()):
         return cleaned_data
 
     def custom_signup(self, request, user):
-        custom_form = super(BaseSignupForm, self)
-        if hasattr(custom_form, "signup") and callable(custom_form.signup):
-            custom_form.signup(request, user)
+        self.signup(request, user)
+
+    def try_save(self, request):
+        """Try and save te user. This can fail in case of a conflict on the
+        email address, in that case we will send an "account already exists"
+        email and return a standard "email verification sent" response.
+        """
+        if self.account_already_exists:
+            # Don't create a new account, only send an email informing the user
+            # that (s)he already has one...
+            email = self.cleaned_data["email"]
+            adapter = get_adapter()
+            adapter.send_account_already_exists_mail(email)
+            user = None
+            resp = adapter.respond_email_verification_sent(request, None)
         else:
-            warnings.warn(
-                "The custom signup form must offer"
-                " a `def signup(self, request, user)` method",
-                DeprecationWarning,
-            )
-            # Historically, it was called .save, but this is confusing
-            # in case of ModelForm
-            custom_form.save(user)
+            user = self.save(request)
+            resp = None
+        return user, resp
 
 
 class SignupForm(BaseSignupForm):
     def __init__(self, *args, **kwargs):
         super(SignupForm, self).__init__(*args, **kwargs)
         self.fields["password1"] = PasswordField(
-            label=_("Password"), autocomplete="new-password"
+            label=_("Password"),
+            autocomplete="new-password",
+            help_text=password_validation.password_validators_help_text_html(),
         )
         if app_settings.SIGNUP_PASSWORD_ENTER_TWICE:
             self.fields["password2"] = PasswordField(
@@ -436,11 +444,8 @@ class SignupForm(BaseSignupForm):
 
     def save(self, request):
         if self.account_already_exists:
-            # Don't create a new acount, only send an email informing the user
-            # that (s)he already has one...
-            self._send_account_already_exists_mail(request)
-            return
-        adapter = get_adapter(request)
+            raise ValueError(self.cleaned_data.get("email"))
+        adapter = get_adapter()
         user = adapter.new_user(request)
         adapter.save_user(request, user, self)
         self.custom_signup(request, user)
@@ -448,23 +453,6 @@ class SignupForm(BaseSignupForm):
         setup_user_email(request, user, [])
         return user
 
-    def _send_account_already_exists_mail(self, request):
-        signup_url = build_absolute_uri(request, reverse("account_signup"))
-        password_reset_url = build_absolute_uri(
-            request, reverse("account_reset_password")
-        )
-        email = self.cleaned_data["email"]
-        context = {
-            "request": request,
-            "current_site": get_current_site(request),
-            "email": email,
-            "signup_url": signup_url,
-            "password_reset_url": password_reset_url,
-        }
-        get_adapter(request).send_mail(
-            "account/email/account_already_exists", email, context
-        )
-
 
 class UserForm(forms.Form):
     def __init__(self, user=None, *args, **kwargs):
@@ -473,26 +461,25 @@ class UserForm(forms.Form):
 
 
 class AddEmailForm(UserForm):
-
     email = forms.EmailField(
-        label=_("E-mail"),
+        label=_("Email"),
         required=True,
         widget=forms.TextInput(
-            attrs={"type": "email", "placeholder": _("E-mail address")}
+            attrs={"type": "email", "placeholder": _("Email address")}
         ),
     )
 
     def clean_email(self):
+        from allauth.account import signals
+
         value = self.cleaned_data["email"]
-        value = get_adapter().clean_email(value)
+        adapter = get_adapter()
+        value = adapter.clean_email(value)
         errors = {
             "this_account": _(
-                "This e-mail address is already associated with this account."
+                "This email address is already associated with this account."
             ),
-            "different_account": _(
-                "This e-mail address is already associated with another account."
-            ),
-            "max_email_addresses": _("You cannot add more than %d e-mail addresses."),
+            "max_email_addresses": _("You cannot add more than %d email addresses."),
         }
         users = filter_users_by_email(value)
         on_this_account = [u for u in users if u.pk == self.user.pk]
@@ -500,26 +487,42 @@ class AddEmailForm(UserForm):
 
         if on_this_account:
             raise forms.ValidationError(errors["this_account"])
-        if on_diff_account and app_settings.UNIQUE_EMAIL:
-            raise forms.ValidationError(errors["different_account"])
+        if (
+            on_diff_account
+            and app_settings.PREVENT_ENUMERATION != "strict"
+            and app_settings.UNIQUE_EMAIL
+        ):
+            raise forms.ValidationError(adapter.error_messages["email_taken"])
         if not EmailAddress.objects.can_add_email(self.user):
             raise forms.ValidationError(
                 errors["max_email_addresses"] % app_settings.MAX_EMAIL_ADDRESSES
             )
+
+        signals._add_email.send(
+            sender=self.user.__class__,
+            email=value,
+            user=self.user,
+        )
         return value
 
     def save(self, request):
+        if app_settings.CHANGE_EMAIL:
+            return EmailAddress.objects.add_new_email(
+                request, self.user, self.cleaned_data["email"]
+            )
         return EmailAddress.objects.add_email(
             request, self.user, self.cleaned_data["email"], confirm=True
         )
 
 
 class ChangePasswordForm(PasswordVerificationMixin, UserForm):
-
     oldpassword = PasswordField(
         label=_("Current Password"), autocomplete="current-password"
     )
-    password1 = SetPasswordField(label=_("New Password"))
+    password1 = SetPasswordField(
+        label=_("New Password"),
+        help_text=password_validation.password_validators_help_text_html(),
+    )
     password2 = PasswordField(label=_("New Password (again)"))
 
     def __init__(self, *args, **kwargs):
@@ -536,8 +539,10 @@ class ChangePasswordForm(PasswordVerificationMixin, UserForm):
 
 
 class SetPasswordForm(PasswordVerificationMixin, UserForm):
-
-    password1 = SetPasswordField(label=_("Password"))
+    password1 = SetPasswordField(
+        label=_("Password"),
+        help_text=password_validation.password_validators_help_text_html(),
+    )
     password2 = PasswordField(label=_("Password (again)"))
 
     def __init__(self, *args, **kwargs):
@@ -549,14 +554,13 @@ class SetPasswordForm(PasswordVerificationMixin, UserForm):
 
 
 class ResetPasswordForm(forms.Form):
-
     email = forms.EmailField(
-        label=_("E-mail"),
+        label=_("Email"),
         required=True,
         widget=forms.TextInput(
             attrs={
                 "type": "email",
-                "placeholder": _("E-mail address"),
+                "placeholder": _("Email address"),
                 "autocomplete": "email",
             }
         ),
@@ -565,10 +569,10 @@ class ResetPasswordForm(forms.Form):
     def clean_email(self):
         email = self.cleaned_data["email"]
         email = get_adapter().clean_email(email)
-        self.users = filter_users_by_email(email, is_active=True)
+        self.users = filter_users_by_email(email, is_active=True, prefer_verified=True)
         if not self.users and not app_settings.PREVENT_ENUMERATION:
             raise forms.ValidationError(
-                _("The e-mail address is not assigned to any user account")
+                _("The email address is not assigned to any user account")
             )
         return self.cleaned_data["email"]
 
@@ -588,13 +592,12 @@ class ResetPasswordForm(forms.Form):
             "request": request,
             "signup_url": signup_url,
         }
-        get_adapter(request).send_mail("account/email/unknown_account", email, context)
+        get_adapter().send_mail("account/email/unknown_account", email, context)
 
     def _send_password_reset_mail(self, request, email, users, **kwargs):
         token_generator = kwargs.get("token_generator", default_token_generator)
 
         for user in users:
-
             temp_key = token_generator.make_token(user)
 
             # save it to the password reset model
@@ -620,13 +623,10 @@ class ResetPasswordForm(forms.Form):
 
             if app_settings.AUTHENTICATION_METHOD != AuthenticationMethod.EMAIL:
                 context["username"] = user_username(user)
-            get_adapter(request).send_mail(
-                "account/email/password_reset_key", email, context
-            )
+            get_adapter().send_mail("account/email/password_reset_key", email, context)
 
 
 class ResetPasswordKeyForm(PasswordVerificationMixin, forms.Form):
-
     password1 = SetPasswordField(label=_("New Password"))
     password2 = PasswordField(label=_("New Password (again)"))
 
@@ -641,7 +641,6 @@ class ResetPasswordKeyForm(PasswordVerificationMixin, forms.Form):
 
 
 class UserTokenForm(forms.Form):
-
     uidb36 = forms.CharField()
     key = forms.CharField()
 
@@ -676,3 +675,19 @@ class UserTokenForm(forms.Form):
             raise forms.ValidationError(self.error_messages["token_invalid"])
 
         return cleaned_data
+
+
+class ReauthenticateForm(forms.Form):
+    password = PasswordField(label=_("Password"), autocomplete="current-password")
+
+    def __init__(self, *args, **kwargs):
+        self.user = kwargs.pop("user")
+        super().__init__(*args, **kwargs)
+
+    def clean_password(self):
+        password = self.cleaned_data.get("password")
+        if not self.user.check_password(password):
+            raise forms.ValidationError(
+                get_adapter().error_messages["incorrect_password"]
+            )
+        return password
