@@ -1,10 +1,11 @@
 import unicodedata
 from collections import OrderedDict
+from typing import Optional
 
 from django.conf import settings
 from django.contrib import messages
-from django.contrib.auth import update_session_auth_hash
-from django.core.exceptions import FieldDoesNotExist, ValidationError
+from django.contrib.auth import get_user_model, update_session_auth_hash
+from django.core.exceptions import FieldDoesNotExist
 from django.db import models
 from django.db.models import Q
 from django.utils.encoding import force_str
@@ -12,10 +13,10 @@ from django.utils.http import base36_to_int, int_to_base36, urlencode
 
 from allauth.account import app_settings, signals
 from allauth.account.adapter import get_adapter
-from allauth.exceptions import ImmediateHttpResponse
+from allauth.account.models import Login
+from allauth.core.exceptions import ImmediateHttpResponse
 from allauth.utils import (
     get_request_param,
-    get_user_model,
     import_callable,
     valid_email_or_none,
 )
@@ -38,7 +39,7 @@ def get_next_redirect_url(request, redirect_field_name="next"):
     via the request.
     """
     redirect_to = get_request_param(request, redirect_field_name)
-    if not get_adapter(request).is_safe_url(redirect_to):
+    if not get_adapter().is_safe_url(redirect_to):
         redirect_to = None
     return redirect_to
 
@@ -53,9 +54,9 @@ def get_login_redirect_url(request, url=None, redirect_field_name="next", signup
         ret = get_next_redirect_url(request, redirect_field_name=redirect_field_name)
     if not ret:
         if signup:
-            ret = get_adapter(request).get_signup_redirect_url(request)
+            ret = get_adapter().get_signup_redirect_url(request)
         else:
-            ret = get_adapter(request).get_login_redirect_url(request)
+            ret = get_adapter().get_login_redirect_url(request)
     return ret
 
 
@@ -85,7 +86,7 @@ def user_display(user):
     return _user_display_callable(user)
 
 
-def user_field(user, field, *args):
+def user_field(user, field, *args, commit=False):
     """
     Gets or sets (optional) user model fields. No-op if fields do not exist.
     """
@@ -105,19 +106,21 @@ def user_field(user, field, *args):
         if v:
             v = v[0:max_length]
         setattr(user, field, v)
+        if commit:
+            user.save(update_fields=[field])
     else:
         # Getter
         return getattr(user, field)
 
 
-def user_username(user, *args):
+def user_username(user, *args, commit=False):
     if args and not app_settings.PRESERVE_USERNAME_CASING and args[0]:
         args = [args[0].lower()]
     return user_field(user, app_settings.USER_MODEL_USERNAME_FIELD, *args)
 
 
-def user_email(user, *args):
-    return user_field(user, app_settings.USER_MODEL_EMAIL_FIELD, *args)
+def user_email(user, *args, commit=False):
+    return user_field(user, app_settings.USER_MODEL_EMAIL_FIELD, *args, commit=commit)
 
 
 def has_verified_email(user, email=None):
@@ -152,24 +155,56 @@ def perform_login(
     email is essential (during signup), or if it can be skipped (e.g. in
     case email verification is optional and we are only logging in).
     """
+    login = Login(
+        user=user,
+        email_verification=email_verification,
+        redirect_url=redirect_url,
+        signal_kwargs=signal_kwargs,
+        signup=signup,
+        email=email,
+    )
+    return _perform_login(request, login)
+
+
+def _perform_login(request, login):
     # Local users are stopped due to form validation checking
     # is_active, yet, adapter methods could toy with is_active in a
     # `user_signed_up` signal. Furthermore, social users should be
     # stopped anyway.
-    adapter = get_adapter(request)
+    adapter = get_adapter()
+    hook_kwargs = _get_login_hook_kwargs(login)
+    response = adapter.pre_login(request, login.user, **hook_kwargs)
+    if response:
+        return response
+    return resume_login(request, login)
+
+
+def _get_login_hook_kwargs(login):
+    """
+    TODO: Just break backwards compatibility and pass only `login` to
+    `pre/post_login()`.
+    """
+    return dict(
+        email_verification=login.email_verification,
+        redirect_url=login.redirect_url,
+        signal_kwargs=login.signal_kwargs,
+        signup=login.signup,
+        email=login.email,
+    )
+
+
+def resume_login(request, login):
+    from allauth.account.stages import LoginStageController
+
+    adapter = get_adapter()
+    ctrl = LoginStageController(request, login)
     try:
-        hook_kwargs = dict(
-            email_verification=email_verification,
-            redirect_url=redirect_url,
-            signal_kwargs=signal_kwargs,
-            signup=signup,
-            email=email,
-        )
-        response = adapter.pre_login(request, user, **hook_kwargs)
+        response = ctrl.handle()
         if response:
             return response
-        adapter.login(request, user)
-        response = adapter.post_login(request, user, **hook_kwargs)
+        adapter.login(request, login.user)
+        hook_kwargs = _get_login_hook_kwargs(login)
+        response = adapter.post_login(request, login.user, **hook_kwargs)
         if response:
             return response
     except ImmediateHttpResponse as e:
@@ -177,6 +212,26 @@ def perform_login(
     return response
 
 
+def unstash_login(request, peek=False):
+    login = None
+    if peek:
+        data = request.session.get("account_login")
+    else:
+        data = request.session.pop("account_login", None)
+    if data is not None:
+        try:
+            login = Login.deserialize(data)
+            request._account_login_accessed = True
+        except ValueError:
+            pass
+    return login
+
+
+def stash_login(request, login):
+    request.session["account_login"] = login.serialize()
+    request._account_login_accessed = True
+
+
 def complete_signup(request, user, email_verification, success_url, signal_kwargs=None):
     if signal_kwargs is None:
         signal_kwargs = {}
@@ -198,12 +253,12 @@ def cleanup_email_addresses(request, addresses):
     Takes a list of EmailAddress instances and cleans it up, making
     sure only valid ones remain, without multiple primaries etc.
 
-    Order is important: e.g. if multiple primary e-mail addresses
+    Order is important: e.g. if multiple primary email addresses
     exist, the first one encountered will be kept as primary.
     """
     from .models import EmailAddress
 
-    adapter = get_adapter(request)
+    adapter = get_adapter()
     # Let's group by `email`
     e2a = OrderedDict()  # maps email to EmailAddress
     primary_addresses = []
@@ -217,8 +272,18 @@ def cleanup_email_addresses(request, addresses):
         # ... and non-conflicting ones...
         if (
             app_settings.UNIQUE_EMAIL
-            and EmailAddress.objects.filter(email__iexact=email).exists()
+            and app_settings.PREVENT_ENUMERATION != "strict"
+            and EmailAddress.objects.lookup([email])
         ):
+            # Email address already exists.
+            continue
+        if (
+            app_settings.UNIQUE_EMAIL
+            and app_settings.PREVENT_ENUMERATION == "strict"
+            and address.verified
+            and EmailAddress.objects.is_verified(email)
+        ):
+            # Email address already exists, and is verified as well.
             continue
         a = e2a.get(email.lower())
         if a:
@@ -265,8 +330,8 @@ def setup_user_email(request, user, addresses):
 
     assert not EmailAddress.objects.filter(user=user).exists()
     priority_addresses = []
-    # Is there a stashed e-mail?
-    adapter = get_adapter(request)
+    # Is there a stashed email?
+    adapter = get_adapter()
     stashed_email = adapter.unstash_verified_email(request)
     if stashed_email:
         priority_addresses.append(
@@ -292,10 +357,10 @@ def setup_user_email(request, user, addresses):
 
 def send_email_confirmation(request, user, signup=False, email=None):
     """
-    E-mail verification mails are sent:
+    Email verification mails are sent:
     a) Explicitly: when a user signs up
     b) Implicitly: when a user attempts to log in using an unverified
-    e-mail while EMAIL_VERIFICATION is mandatory.
+    email while EMAIL_VERIFICATION is mandatory.
 
     Especially in case of b), we want to limit the number of mails
     sent (consider a user retrying a few times), which is why there is
@@ -304,7 +369,7 @@ def send_email_confirmation(request, user, signup=False, email=None):
     """
     from .models import EmailAddress
 
-    adapter = get_adapter(request)
+    adapter = get_adapter()
 
     if not email:
         email = user_email(user)
@@ -331,7 +396,7 @@ def send_email_confirmation(request, user, signup=False, email=None):
                 request,
                 messages.INFO,
                 "account/messages/email_confirmation_sent.txt",
-                {"email": email},
+                {"email": email, "login": not signup, "signup": signup},
             )
     if signup:
         adapter.stash_user(request, user_pk_to_url_str(user))
@@ -352,12 +417,6 @@ def sync_user_email_addresses(user):
         email
         and not EmailAddress.objects.filter(user=user, email__iexact=email).exists()
     ):
-        if (
-            app_settings.UNIQUE_EMAIL
-            and EmailAddress.objects.filter(email__iexact=email).exists()
-        ):
-            # Bail out
-            return
         # get_or_create() to gracefully handle races
         EmailAddress.objects.get_or_create(
             user=user, email=email, defaults={"primary": False, "verified": False}
@@ -384,24 +443,37 @@ def filter_users_by_username(*username):
     return ret
 
 
-def filter_users_by_email(email, is_active=None):
+def filter_users_by_email(email, is_active=None, prefer_verified=False):
     """Return list of users by email address
 
     Typically one, at most just a few in length.  First we look through
     EmailAddress table, than customisable User model table. Add results
     together avoiding SQL joins and deduplicate.
+
+    `prefer_verified`: When looking up users by email, there can be cases where
+    users with verified email addresses are preferable above users who did not
+    verify their email address. The password reset is such a use case -- if
+    there is a user with a verified email than that user should be returned, not
+    one of the other users.
     """
     from .models import EmailAddress
 
     User = get_user_model()
-    mails = EmailAddress.objects.filter(email__iexact=email)
+    mails = EmailAddress.objects.filter(email__iexact=email).prefetch_related("user")
     if is_active is not None:
         mails = mails.filter(user__is_active=is_active)
+    mails = list(mails)
+    is_verified = False
+    if prefer_verified:
+        verified_mails = list(filter(lambda e: e.verified, mails))
+        if verified_mails:
+            mails = verified_mails
+            is_verified = True
     users = []
-    for e in mails.prefetch_related("user"):
+    for e in mails:
         if _unicode_ci_compare(e.email, email):
             users.append(e.user)
-    if app_settings.USER_MODEL_EMAIL_FIELD:
+    if app_settings.USER_MODEL_EMAIL_FIELD and not is_verified:
         q_dict = {app_settings.USER_MODEL_EMAIL_FIELD + "__iexact": email}
         user_qs = User.objects.filter(**q_dict)
         if is_active is not None:
@@ -426,31 +498,65 @@ def user_pk_to_url_str(user):
     This should return a string.
     """
     User = get_user_model()
-    if issubclass(type(User._meta.pk), models.UUIDField):
+    pk_field_class = type(User._meta.pk)
+    if issubclass(pk_field_class, models.UUIDField):
         if isinstance(user.pk, str):
             return user.pk
         return user.pk.hex
-
-    ret = user.pk
-    if isinstance(ret, int):
-        ret = int_to_base36(user.pk)
-    return str(ret)
+    elif issubclass(pk_field_class, models.IntegerField):
+        return int_to_base36(int(user.pk))
+    return str(user.pk)
 
 
-def url_str_to_user_pk(s):
+def url_str_to_user_pk(pk_str):
     User = get_user_model()
-    # TODO: Ugh, isn't there a cleaner way to determine whether or not
-    # the PK is a str-like field?
     remote_field = getattr(User._meta.pk, "remote_field", None)
     if remote_field and getattr(remote_field, "to", None):
         pk_field = User._meta.pk.remote_field.to._meta.pk
     else:
         pk_field = User._meta.pk
-    if issubclass(type(pk_field), models.UUIDField):
-        return pk_field.to_python(s)
-    try:
-        pk_field.to_python("a")
-        pk = s
-    except ValidationError:
-        pk = base36_to_int(s)
+    pk_field_class = type(pk_field)
+    if issubclass(pk_field_class, models.IntegerField):
+        pk = base36_to_int(pk_str)
+        # always call to_python() -- because there are fields like HashidField
+        # that derive from IntegerField.
+        pk = pk_field.to_python(pk)
+    else:
+        pk = pk_field.to_python(pk_str)
     return pk
+
+
+def assess_unique_email(email) -> Optional[bool]:
+    """
+    True -- email is unique
+    False -- email is already in use
+    None -- email is in use, but we should hide that using email verification.
+    """
+    if not filter_users_by_email(email):
+        # All good.
+        return True
+    elif not app_settings.PREVENT_ENUMERATION:
+        # Fail right away.
+        return False
+    elif (
+        app_settings.EMAIL_VERIFICATION
+        == app_settings.EmailVerificationMethod.MANDATORY
+    ):
+        # In case of mandatory verification and enumeration prevention,
+        # we can avoid creating a new account with the same (unverified)
+        # email address, because we are going to send an email anyway.
+        assert app_settings.PREVENT_ENUMERATION
+        return None
+    elif app_settings.PREVENT_ENUMERATION == "strict":
+        # We're going to be strict on enumeration prevention, and allow for
+        # this email address to pass even though it already exists. In this
+        # scenario, you can signup multiple times using the same email
+        # address resulting in multiple accounts with an unverified email.
+        return True
+    else:
+        assert app_settings.PREVENT_ENUMERATION is True
+        # Conflict. We're supposed to prevent enumeration, but we can't
+        # because that means letting the user in, while emails are required
+        # to be unique. In this case, uniqueness takes precedence over
+        # enumeration prevention.
+        return False
