@@ -1,6 +1,9 @@
 from django.contrib import messages
+from django.contrib.auth import REDIRECT_FIELD_NAME
 from django.contrib.auth.decorators import login_required
 from django.contrib.sites.shortcuts import get_current_site
+from django.core.validators import validate_email
+from django.forms import ValidationError
 from django.http import (
     Http404,
     HttpResponsePermanentRedirect,
@@ -9,29 +12,35 @@ from django.http import (
 from django.shortcuts import redirect
 from django.urls import reverse, reverse_lazy
 from django.utils.decorators import method_decorator
+from django.views.decorators.cache import never_cache
 from django.views.decorators.debug import sensitive_post_parameters
 from django.views.generic.base import TemplateResponseMixin, TemplateView, View
 from django.views.generic.edit import FormView
 
-from allauth import ratelimit
-from allauth.decorators import rate_limit
-from allauth.exceptions import ImmediateHttpResponse
-from allauth.utils import get_form_class, get_request_param
-
-from . import app_settings, signals
-from .adapter import get_adapter
-from .forms import (
+from allauth.account import app_settings, signals
+from allauth.account.adapter import get_adapter
+from allauth.account.decorators import reauthentication_required
+from allauth.account.forms import (
     AddEmailForm,
     ChangePasswordForm,
     LoginForm,
+    ReauthenticateForm,
     ResetPasswordForm,
     ResetPasswordKeyForm,
     SetPasswordForm,
     SignupForm,
     UserTokenForm,
 )
-from .models import EmailAddress, EmailConfirmation, EmailConfirmationHMAC
-from .utils import (
+from allauth.account.models import (
+    EmailAddress,
+    EmailConfirmation,
+    EmailConfirmationHMAC,
+)
+from allauth.account.reauthentication import (
+    record_authentication,
+    resume_request,
+)
+from allauth.account.utils import (
     complete_signup,
     get_login_redirect_url,
     get_next_redirect_url,
@@ -42,6 +51,10 @@ from .utils import (
     sync_user_email_addresses,
     url_str_to_user_pk,
 )
+from allauth.core import ratelimit
+from allauth.core.exceptions import ImmediateHttpResponse
+from allauth.decorators import rate_limit
+from allauth.utils import get_form_class, get_request_param
 
 
 INTERNAL_RESET_SESSION_KEY = "_password_reset_key"
@@ -53,7 +66,7 @@ sensitive_post_parameters_m = method_decorator(
 
 
 def _ajax_response(request, response, form=None, data=None):
-    adapter = get_adapter(request)
+    adapter = get_adapter()
     if adapter.is_ajax(request):
         if isinstance(response, HttpResponseRedirect) or isinstance(
             response, HttpResponsePermanentRedirect
@@ -142,9 +155,10 @@ class LoginView(
     form_class = LoginForm
     template_name = "account/login." + app_settings.TEMPLATE_EXTENSION
     success_url = None
-    redirect_field_name = "next"
+    redirect_field_name = REDIRECT_FIELD_NAME
 
     @sensitive_post_parameters_m
+    @method_decorator(never_cache)
     def dispatch(self, request, *args, **kwargs):
         return super(LoginView, self).dispatch(request, *args, **kwargs)
 
@@ -226,10 +240,11 @@ class SignupView(
 ):
     template_name = "account/signup." + app_settings.TEMPLATE_EXTENSION
     form_class = SignupForm
-    redirect_field_name = "next"
+    redirect_field_name = REDIRECT_FIELD_NAME
     success_url = None
 
     @sensitive_post_parameters_m
+    @method_decorator(never_cache)
     def dispatch(self, request, *args, **kwargs):
         return super(SignupView, self).dispatch(request, *args, **kwargs)
 
@@ -245,11 +260,9 @@ class SignupView(
         return ret
 
     def form_valid(self, form):
-        self.user = form.save(self.request)
-        if not self.user:
-            return get_adapter(self.request).respond_email_verification_sent(
-                self.request, None
-            )
+        self.user, resp = form.try_save(self.request)
+        if resp:
+            return resp
         try:
             return complete_signup(
                 self.request,
@@ -286,12 +299,24 @@ class SignupView(
         )
         return ret
 
+    def get_initial(self):
+        initial = super().get_initial()
+        email = self.request.GET.get("email")
+        if email:
+            try:
+                validate_email(email)
+            except ValidationError:
+                return initial
+            initial["email"] = email
+            if app_settings.SIGNUP_EMAIL_ENTER_TWICE:
+                initial["email2"] = email
+        return initial
+
 
 signup = SignupView.as_view()
 
 
 class ConfirmEmailView(TemplateResponseMixin, LogoutFunctionalityMixin, View):
-
     template_name = "account/email_confirm." + app_settings.TEMPLATE_EXTENSION
 
     def get(self, *args, **kwargs):
@@ -306,7 +331,15 @@ class ConfirmEmailView(TemplateResponseMixin, LogoutFunctionalityMixin, View):
 
     def post(self, *args, **kwargs):
         self.object = confirmation = self.get_object()
-        confirmation.confirm(self.request)
+        email_address = confirmation.confirm(self.request)
+        if not email_address:
+            get_adapter(self.request).add_message(
+                self.request,
+                messages.ERROR,
+                "account/messages/email_confirmation_failed.txt",
+                {"email": confirmation.email_address.email},
+            )
+            return self.respond(False)
 
         # In the event someone clicks on an email confirmation link
         # for one account while logged into another account,
@@ -333,6 +366,9 @@ class ConfirmEmailView(TemplateResponseMixin, LogoutFunctionalityMixin, View):
         # user = confirmation.email_address.user
         # user.is_active = True
         # user.save()
+        return self.respond(True)
+
+    def respond(self, success):
         redirect_url = self.get_redirect_url()
         if not redirect_url:
             ctx = self.get_context_data()
@@ -398,9 +434,17 @@ class ConfirmEmailView(TemplateResponseMixin, LogoutFunctionalityMixin, View):
 
     def get_context_data(self, **kwargs):
         ctx = kwargs
-        ctx["confirmation"] = self.object
         site = get_current_site(self.request)
-        ctx.update({"site": site})
+        ctx.update(
+            {
+                "site": site,
+                "confirmation": self.object,
+                "can_confirm": self.object
+                and self.object.email_address.can_set_verified(),
+            }
+        )
+        if self.object:
+            ctx["email"] = self.object.email_address.email
         return ctx
 
     def get_redirect_url(self):
@@ -413,8 +457,16 @@ confirm_email = ConfirmEmailView.as_view()
 
 
 @method_decorator(rate_limit(action="manage_email"), name="dispatch")
+@method_decorator(
+    reauthentication_required(
+        allow_get=True, enabled=lambda request: app_settings.REAUTHENTICATION_REQUIRED
+    ),
+    name="dispatch",
+)
 class EmailView(AjaxCapableProcessFormViewMixin, FormView):
-    template_name = "account/email." + app_settings.TEMPLATE_EXTENSION
+    template_name = (
+        "account/email_change." if app_settings.CHANGE_EMAIL else "account/email."
+    ) + app_settings.TEMPLATE_EXTENSION
     form_class = AddEmailForm
     success_url = reverse_lazy("account_email")
 
@@ -469,6 +521,10 @@ class EmailView(AjaxCapableProcessFormViewMixin, FormView):
 
     def _get_email_address(self, request):
         email = request.POST["email"]
+        try:
+            validate_email(email)
+        except ValidationError:
+            return None
         try:
             return EmailAddress.objects.get_for_user(user=request.user, email=email)
         except EmailAddress.DoesNotExist:
@@ -484,22 +540,23 @@ class EmailView(AjaxCapableProcessFormViewMixin, FormView):
     def _action_remove(self, request, *args, **kwargs):
         email_address = self._get_email_address(request)
         if email_address:
-            if email_address.primary:
-                get_adapter(request).add_message(
+            adapter = get_adapter()
+            if not adapter.can_delete_email(email_address):
+                adapter.add_message(
                     request,
                     messages.ERROR,
                     "account/messages/cannot_delete_primary_email.txt",
                     {"email": email_address.email},
                 )
             else:
-                email_address.delete()
+                email_address.remove()
                 signals.email_removed.send(
                     sender=request.user.__class__,
                     request=request,
                     user=request.user,
                     email_address=email_address,
                 )
-                get_adapter(request).add_message(
+                adapter.add_message(
                     request,
                     messages.SUCCESS,
                     "account/messages/email_deleted.txt",
@@ -520,7 +577,7 @@ class EmailView(AjaxCapableProcessFormViewMixin, FormView):
                     user=request.user, verified=True
                 ).exists()
             ):
-                get_adapter(request).add_message(
+                get_adapter().add_message(
                     request,
                     messages.ERROR,
                     "account/messages/unverified_primary_email.txt",
@@ -535,7 +592,7 @@ class EmailView(AjaxCapableProcessFormViewMixin, FormView):
                 except EmailAddress.DoesNotExist:
                     from_email_address = None
                 email_address.set_as_primary()
-                get_adapter(request).add_message(
+                get_adapter().add_message(
                     request,
                     messages.SUCCESS,
                     "account/messages/primary_email_set.txt",
@@ -551,15 +608,38 @@ class EmailView(AjaxCapableProcessFormViewMixin, FormView):
 
     def get_context_data(self, **kwargs):
         ret = super(EmailView, self).get_context_data(**kwargs)
-        # NOTE: For backwards compatibility
-        ret["add_email_form"] = ret.get("form")
-        # (end NOTE)
-        ret["can_add_email"] = EmailAddress.objects.can_add_email(self.request.user)
+        emails = list(
+            EmailAddress.objects.filter(user=self.request.user).order_by("email")
+        )
+        ret.update(
+            {
+                "emailaddresses": emails,
+                "emailaddress_radios": [
+                    {
+                        "id": f"email_radio_{i}",
+                        "checked": email.primary or len(emails) == 1,
+                        "emailaddress": email,
+                    }
+                    for i, email in enumerate(emails)
+                ],
+                "add_email_form": ret.get("form"),
+                "can_add_email": EmailAddress.objects.can_add_email(self.request.user),
+            }
+        )
+        if app_settings.CHANGE_EMAIL:
+            ret.update(
+                {
+                    "new_emailaddress": EmailAddress.objects.get_new(self.request.user),
+                    "current_emailaddress": EmailAddress.objects.get_verified(
+                        self.request.user
+                    ),
+                }
+            )
         return ret
 
     def get_ajax_data(self):
         data = []
-        for emailaddress in self.request.user.emailaddress_set.all():
+        for emailaddress in self.request.user.emailaddress_set.all().order_by("pk"):
             data.append(
                 {
                     "id": emailaddress.pk,
@@ -588,7 +668,13 @@ class PasswordChangeView(AjaxCapableProcessFormViewMixin, FormView):
         return super(PasswordChangeView, self).dispatch(request, *args, **kwargs)
 
     def render_to_response(self, context, **response_kwargs):
-        if not self.request.user.has_usable_password():
+        if self.request.user.is_anonymous:
+            # We end up here when `ACCOUNT_LOGOUT_ON_PASSWORD_CHANGE = True`.
+            redirect_url = get_adapter(self.request).get_logout_redirect_url(
+                self.request
+            )
+            return HttpResponseRedirect(redirect_url)
+        elif not self.request.user.has_usable_password():
             return HttpResponseRedirect(reverse("account_set_password"))
         return super(PasswordChangeView, self).render_to_response(
             context, **response_kwargs
@@ -684,7 +770,7 @@ class PasswordResetView(AjaxCapableProcessFormViewMixin, FormView):
     template_name = "account/password_reset." + app_settings.TEMPLATE_EXTENSION
     form_class = ResetPasswordForm
     success_url = reverse_lazy("account_reset_password_done")
-    redirect_field_name = "next"
+    redirect_field_name = REDIRECT_FIELD_NAME
 
     def get_form_class(self):
         return get_form_class(app_settings.FORMS, "reset_password", self.form_class)
@@ -740,10 +826,13 @@ class PasswordResetFromKeyView(
         self.request = request
         self.key = key
 
+        user_token_form_class = get_form_class(
+            app_settings.FORMS, "user_token", UserTokenForm
+        )
         if self.key == self.reset_url_key:
             self.key = self.request.session.get(INTERNAL_RESET_SESSION_KEY, "")
             # (Ab)using forms here to be able to handle errors in XHR #890
-            token_form = UserTokenForm(data={"uidb36": uidb36, "key": self.key})
+            token_form = user_token_form_class(data={"uidb36": uidb36, "key": self.key})
             if token_form.is_valid():
                 self.reset_user = token_form.reset_user
 
@@ -761,7 +850,7 @@ class PasswordResetFromKeyView(
                     request, uidb36, self.key, **kwargs
                 )
         else:
-            token_form = UserTokenForm(data={"uidb36": uidb36, "key": self.key})
+            token_form = user_token_form_class(data={"uidb36": uidb36, "key": self.key})
             if token_form.is_valid():
                 # Store the key in the session and redirect to the
                 # password reset form at a URL without the key. That
@@ -838,9 +927,8 @@ password_reset_from_key_done = PasswordResetFromKeyDoneView.as_view()
 
 
 class LogoutView(TemplateResponseMixin, LogoutFunctionalityMixin, View):
-
     template_name = "account/logout." + app_settings.TEMPLATE_EXTENSION
-    redirect_field_name = "next"
+    redirect_field_name = REDIRECT_FIELD_NAME
 
     def get(self, *args, **kwargs):
         if app_settings.LOGOUT_ON_GET:
@@ -891,3 +979,54 @@ class EmailVerificationSentView(TemplateView):
 
 
 email_verification_sent = EmailVerificationSentView.as_view()
+
+
+class ReauthenticateView(FormView):
+    form_class = ReauthenticateForm
+    template_name = "account/reauthenticate." + app_settings.TEMPLATE_EXTENSION
+    redirect_field_name = REDIRECT_FIELD_NAME
+
+    def dispatch(self, request, *args, **kwargs):
+        r429 = ratelimit.consume_or_429(
+            self.request,
+            action="reauthenticate",
+            user=self.request.user,
+        )
+        if r429:
+            return r429
+        return super().dispatch(request, *args, **kwargs)
+
+    def get_form_class(self):
+        return get_form_class(app_settings.FORMS, "reauthenticate", self.form_class)
+
+    def get_success_url(self):
+        url = get_next_redirect_url(self.request, self.redirect_field_name)
+        if not url:
+            url = get_adapter(self.request).get_login_redirect_url(self.request)
+        return url
+
+    def get_form_kwargs(self):
+        ret = super().get_form_kwargs()
+        ret["user"] = self.request.user
+        return ret
+
+    def form_valid(self, form):
+        record_authentication(self.request, self.request.user)
+        response = resume_request(self.request)
+        if response:
+            return response
+        return super().form_valid(form)
+
+    def get_context_data(self, **kwargs):
+        ret = super(ReauthenticateView, self).get_context_data(**kwargs)
+        redirect_field_value = get_request_param(self.request, self.redirect_field_name)
+        ret.update(
+            {
+                "redirect_field_name": self.redirect_field_name,
+                "redirect_field_value": redirect_field_value,
+            }
+        )
+        return ret
+
+
+reauthenticate = login_required(ReauthenticateView.as_view())
