@@ -1,17 +1,24 @@
 from __future__ import absolute_import
 
-from django.core.exceptions import ValidationError
+import warnings
+
+from django.core.exceptions import (
+    ImproperlyConfigured,
+    MultipleObjectsReturned,
+    ValidationError,
+)
+from django.db.models import Q
 from django.urls import reverse
 from django.utils.translation import gettext_lazy as _
 
-from ..account import app_settings as account_settings
+from allauth.core import context
+
 from ..account.adapter import get_adapter as get_account_adapter
 from ..account.app_settings import EmailVerificationMethod
 from ..account.models import EmailAddress
 from ..account.utils import user_email, user_field, user_username
 from ..utils import (
     deserialize_instance,
-    email_address_exists,
     import_attribute,
     serialize_instance,
     valid_email_or_none,
@@ -20,17 +27,18 @@ from . import app_settings
 
 
 class DefaultSocialAccountAdapter(object):
-
     error_messages = {
         "email_taken": _(
-            "An account already exists with this e-mail address."
+            "An account already exists with this email address."
             " Please sign in to that account first, then connect"
             " your %s account."
         )
     }
 
     def __init__(self, request=None):
-        self.request = request
+        # Explicitly passing `request` is deprecated, just use:
+        # `allauth.core.context.request`.
+        self.request = context.request
 
     def pre_social_login(self, request, sociallogin):
         """
@@ -134,35 +142,12 @@ class DefaultSocialAccountAdapter(object):
                     user=account.user, verified=True
                 ).exists():
                     raise ValidationError(
-                        _("Your account has no verified e-mail address.")
+                        _("Your account has no verified email address.")
                     )
 
     def is_auto_signup_allowed(self, request, sociallogin):
         # If email is specified, check for duplicate and if so, no auto signup.
         auto_signup = app_settings.AUTO_SIGNUP
-        if auto_signup:
-            email = user_email(sociallogin.user)
-            # Let's check if auto_signup is really possible...
-            if email:
-                if account_settings.UNIQUE_EMAIL:
-                    if email_address_exists(email):
-                        # Oops, another user already has this address.
-                        # We cannot simply connect this social account
-                        # to the existing user. Reason is that the
-                        # email address may not be verified, meaning,
-                        # the user may be a hacker that has added your
-                        # email address to their account in the hope
-                        # that you fall in their trap.  We cannot
-                        # check on 'email_address.verified' either,
-                        # because 'email_address' is not guaranteed to
-                        # be verified.
-                        auto_signup = False
-                        # FIXME: We redirect to signup form -- user will
-                        # see email address conflict only after posting
-                        # whereas we detected it here already.
-            elif app_settings.EMAIL_REQUIRED:
-                # Nope, email is required and we don't have it yet...
-                auto_signup = False
         return auto_signup
 
     def is_open_for_signup(self, request, sociallogin):
@@ -190,18 +175,120 @@ class DefaultSocialAccountAdapter(object):
     def serialize_instance(self, instance):
         return serialize_instance(instance)
 
-    def get_app(self, request, provider, config=None):
+    def list_providers(self, request):
+        from allauth.socialaccount.providers import registry
+
+        ret = []
+        provider_classes = registry.get_class_list()
+        apps = self.list_apps(request)
+        apps_map = {}
+        for app in apps:
+            apps_map.setdefault(app.provider, []).append(app)
+        for provider_class in provider_classes:
+            provider_apps = apps_map.get(provider_class.id, [])
+            if not provider_apps:
+                if provider_class.uses_apps:
+                    continue
+                provider_apps = [None]
+            for app in provider_apps:
+                provider = provider_class(request=request, app=app)
+                ret.append(provider)
+        return ret
+
+    def get_provider(self, request, provider):
+        """Looks up a `provider`, supporting subproviders by looking up by
+        `provider_id`.
+        """
+        from allauth.socialaccount.providers import registry
+
+        provider_class = registry.get_class(provider)
+        if provider_class is None or provider_class.uses_apps:
+            app = self.get_app(request, provider=provider)
+            if not provider_class:
+                # In this case, the `provider` argument passed was a
+                # `provider_id`.
+                provider_class = registry.get_class(app.provider)
+            if not provider_class:
+                raise ImproperlyConfigured(f"unknown provider: {app.provider}")
+            return provider_class(request, app=app)
+        elif provider_class:
+            assert not provider_class.uses_apps
+            return provider_class(request, app=None)
+        else:
+            raise ImproperlyConfigured(f"unknown provider: {app.provider}")
+
+    def list_apps(self, request, provider=None, client_id=None):
+        """SocialApp's can be setup in the database, or, via
+        `settings.SOCIALACCOUNT_PROVIDERS`.  This methods returns a uniform list
+        of all known apps matching the specified criteria, and blends both
+        (db/settings) sources of data.
+        """
         # NOTE: Avoid loading models at top due to registry boot...
         from allauth.socialaccount.models import SocialApp
 
-        config = config or app_settings.PROVIDERS.get(provider, {}).get("APP")
-        if config:
-            app = SocialApp(provider=provider)
-            for field in ["client_id", "secret", "key", "certificate_key"]:
-                setattr(app, field, config.get(field))
-        else:
-            app = SocialApp.objects.get_current(provider, request)
-        return app
+        # Map provider to the list of apps.
+        provider_to_apps = {}
+
+        # First, populate it with the DB backed apps.
+        db_apps = SocialApp.objects.on_site(request)
+        if provider:
+            db_apps = db_apps.filter(Q(provider=provider) | Q(provider_id=provider))
+        if client_id:
+            db_apps = db_apps.filter(client_id=client_id)
+        for app in db_apps:
+            apps = provider_to_apps.setdefault(app.provider, [])
+            apps.append(app)
+
+        # Then, extend it with the settings backed apps.
+        for p, pcfg in app_settings.PROVIDERS.items():
+            app_configs = pcfg.get("APPS")
+            if app_configs is None:
+                app_config = pcfg.get("APP")
+                if app_config is None:
+                    continue
+                app_configs = [app_config]
+
+            apps = provider_to_apps.setdefault(p, [])
+            for config in app_configs:
+                app = SocialApp(provider=p)
+                for field in [
+                    "name",
+                    "provider_id",
+                    "client_id",
+                    "secret",
+                    "key",
+                    "settings",
+                ]:
+                    if field in config:
+                        setattr(app, field, config[field])
+                if "certificate_key" in config:
+                    warnings.warn("'certificate_key' should be moved into app.settings")
+                    app.settings["certificate_key"] = config["certificate_key"]
+                if client_id and app.client_id != client_id:
+                    continue
+                if (
+                    provider
+                    and app.provider_id != provider
+                    and app.provider != provider
+                ):
+                    continue
+                apps.append(app)
+
+        # Flatten the list of apps.
+        apps = []
+        for provider_apps in provider_to_apps.values():
+            apps.extend(provider_apps)
+        return apps
+
+    def get_app(self, request, provider, client_id=None):
+        from allauth.socialaccount.models import SocialApp
+
+        apps = self.list_apps(request, provider=provider, client_id=client_id)
+        if len(apps) > 1:
+            raise MultipleObjectsReturned
+        elif len(apps) == 0:
+            raise SocialApp.DoesNotExist()
+        return apps[0]
 
 
 def get_adapter(request=None):
