@@ -1,8 +1,10 @@
 from __future__ import absolute_import
 
+import logging
+
 from django.urls import reverse
 
-from allauth.socialaccount import providers
+from allauth.socialaccount.adapter import get_adapter
 from allauth.socialaccount.helpers import (
     complete_social_login,
     render_authentication_error,
@@ -19,6 +21,9 @@ from allauth.socialaccount.providers.oauth.client import (
 )
 
 
+logger = logging.getLogger(__name__)
+
+
 class OAuthAdapter(object):
     def __init__(self, request):
         self.request = request
@@ -30,7 +35,9 @@ class OAuthAdapter(object):
         raise NotImplementedError
 
     def get_provider(self):
-        return providers.registry.by_id(self.provider_id, self.request)
+        adapter = get_adapter(self.request)
+        app = adapter.get_app(self.request, provider=self.provider_id)
+        return app.get_provider(self.request)
 
 
 class OAuthView(object):
@@ -46,7 +53,7 @@ class OAuthView(object):
 
     def _get_client(self, request, callback_url):
         provider = self.adapter.get_provider()
-        app = provider.get_app(request)
+        app = provider.app
         scope = " ".join(provider.get_scope(request))
         parameters = {}
         if scope:
@@ -76,6 +83,7 @@ class OAuthLoginView(OAuthLoginMixin, OAuthView):
         try:
             return client.get_redirect(auth_url, auth_params)
         except OAuthError as e:
+            logger.error("OAuth authentication error", exc_info=True)
             return render_authentication_error(
                 request, self.adapter.provider_id, exception=e
             )
@@ -101,7 +109,7 @@ class OAuthCallbackView(OAuthView):
                 error=error,
                 extra_context=extra_context,
             )
-        app = self.adapter.get_provider().get_app(request)
+        app = self.adapter.get_provider().app
         try:
             access_token = client.get_access_token()
             token = SocialToken(
