@@ -1,5 +1,6 @@
 from unittest import expectedFailure
 
+from django.contrib.auth import get_user_model
 from django.test import override_settings
 from django.urls import reverse
 
@@ -7,7 +8,6 @@ from openid.consumer import consumer
 
 from allauth.socialaccount.models import SocialAccount
 from allauth.tests import Mock, TestCase, patch
-from allauth.utils import get_user_model
 
 from . import views
 from .utils import AXAttribute
@@ -33,7 +33,7 @@ class OpenIDTests(TestCase):
         )
         assert "login.yahooapis" in resp["location"]
         with patch(
-            "allauth.socialaccount.providers" ".openid.views._openid_consumer"
+            "allauth.socialaccount.providers.openid.views._openid_consumer"
         ) as consumer_mock:
             client = Mock()
             complete = Mock()
@@ -44,10 +44,10 @@ class OpenIDTests(TestCase):
             complete_response.status = consumer.SUCCESS
             complete_response.identity_url = "http://dummy/john/"
             with patch(
-                "allauth.socialaccount.providers" ".openid.utils.SRegResponse"
+                "allauth.socialaccount.providers.openid.utils.SRegResponse"
             ) as sr_mock:
                 with patch(
-                    "allauth.socialaccount.providers" ".openid.utils.FetchResponse"
+                    "allauth.socialaccount.providers.openid.utils.FetchResponse"
                 ) as fr_mock:
                     sreg_mock = Mock()
                     ax_mock = Mock()
@@ -91,7 +91,7 @@ class OpenIDTests(TestCase):
             )
         assert "login.yahooapis" in resp["location"]
         with patch(
-            "allauth.socialaccount.providers" ".openid.views._openid_consumer"
+            "allauth.socialaccount.providers.openid.views._openid_consumer"
         ) as consumer_mock:
             client = Mock()
             complete = Mock()
@@ -105,10 +105,10 @@ class OpenIDTests(TestCase):
             complete_response.status = consumer.SUCCESS
             complete_response.identity_url = "http://dummy/john/"
             with patch(
-                "allauth.socialaccount.providers" ".openid.utils.SRegResponse"
+                "allauth.socialaccount.providers.openid.utils.SRegResponse"
             ) as sr_mock:
                 with patch(
-                    "allauth.socialaccount.providers" ".openid.utils.FetchResponse"
+                    "allauth.socialaccount.providers.openid.utils.FetchResponse"
                 ) as fr_mock:
                     sreg_mock = Mock()
                     ax_mock = Mock()
